import { fileURLToPath as fpath } from "node:url";
import path from "node:path";
import http from "node:http";
import express, { Router } from "express";
import { generateApi } from "./express/ApiGenerator.js";
import clientIpMiddleware from "./express/middlewares/client-ip-middleware.js";
import vitalsRoute from "./express/routes/vitals-route.js";
import { clientErrorRoute } from "./express/ClientErrors.js";
import { baseRedirectRoute, staticRoute } from "./exports/index.express.js";
import { WARN } from "./exports/index.logging.js";
import autoBind from "auto-bind";
import "express-async-errors";
const defaultPort = 52000;
const InternalVixConfigure = Symbol("internal-vix-configure");
export class VixCore {
    constructor(title, port) {
        this.title = title;
        this.port = port ?? defaultPort;
        this.app = express();
        this.httpServer = http.createServer(this.app);
    }
    async preConfigure() { }
    [InternalVixConfigure] = async () => { };
    async postConfigure() { }
    async start() {
        const vix = this;
        return new Promise(async function (res, _rej) {
            await vix.preConfigure();
            await vix[InternalVixConfigure]();
            await vix.configureServices();
            await vix.postConfigure();
            await vix.startServices();
            vix.httpServer.listen(vix.port, function onUp() {
                vix.onStart();
                res();
            });
        });
    }
}
export class VixpressService {
    app;
    httpServer;
    constructor(app, httpServer) {
        this.app = app;
        this.httpServer = httpServer;
    }
    async configureService() { }
    async startService() { }
}
export class Vix extends VixCore {
    constructor(title, port, services = new Map()) {
        super(title, port);
        this.serviceClasses = services;
        this.initialized = false;
    }
    [InternalVixConfigure] = async () => {
        this.services = new Map();
        for (const [s, Service] of this.serviceClasses.entries()) {
            const service = new Service(this.app, this.httpServer);
            this.app.set(s, service);
            this.services.set(s, service);
        }
    };
    async configureServices() {
        if (this.initialized)
            throw Error("Vix can only be configured once!");
        const configPromises = [];
        for (const svc of this.services.values())
            configPromises.push(svc.configureService());
        this.initialized = true;
        await Promise.all(configPromises);
    }
    setService(serviceKey, service) {
        this.serviceClasses.set(serviceKey, service);
    }
    async startServices() {
        for (const svc of this.services.values())
            await svc.startService();
    }
    async onStart() { }
}
export class Vixpress extends Vix {
    static RouterService = "router";
    static Tip = Error(`Set the router in the constructor via setRouter`);
    constructor(title, port) {
        super(title, port);
        this.routerReady = false;
    }
    setRouter(RouterService) {
        this.routerService = new RouterService(this.app, this.httpServer);
        this.app.set(Vixpress.RouterService, this.routerService);
    }
    async configureServices() {
        if (!this.routerService)
            throw Vixpress.Tip;
        if (!this.routerReady)
            await this.routerService.configureService();
        return super.configureServices();
    }
    async generateApi() {
        if (!this.routerService)
            throw Vixpress.Tip;
        if (!this.routerReady)
            await this.routerService.configureService();
        return generateApi(this.app);
    }
    setService(serviceKey, Service) {
        if (serviceKey === Vixpress.RouterService)
            throw Vixpress.Tip;
        super.setService(serviceKey, Service);
    }
}
export class VixpressController {
    constructor(app) {
        this.app = app;
        autoBind(this);
    }
}
export class VixpressRoute {
    constructor(app) {
        this.app = app;
        this.router = this.buildRouter();
    }
    buildRouter() {
        return Router({ mergeParams: true });
    }
    useController(Controller) {
        const controller = new Controller(this.app);
        return controller;
    }
}
export class VixpressRouter extends VixpressService {
    constructor(app, httpServer) {
        super(app, httpServer);
        this.router = this.buildRouter();
    }
    buildRouter() {
        return Router({ mergeParams: true });
    }
    async useRoute(route, ControllerRoute) {
        const controller = new ControllerRoute(this.app);
        await controller.configureRoutes(controller.router);
        this.router.use(route, controller.router);
    }
    async configureService() {
        // This function is called by Vixpress to configure the app
        this.configureCore();
        this.configureVite();
        await this.configureRoutes(this.router);
        this.configureErrorHandling();
        this.app.use(this.router);
    }
    configureCore() {
        // Setup IP forwarding & vitals
        this.router.use(clientIpMiddleware());
        this.router.use(vitalsRoute);
    }
    configureVite() {
        // Setup statics for Vite assets
        if (!this.routerImportUrl)
            throw Error(`Please set the routerImportUrl by defining it at the top of your VixpressRouter like so: 'protected routerImportUrl = import.meta.url;'`);
        if (!this.baseUrl)
            WARN("WARN", `Please set the baseUrl by defining it at the top of your VixpressRouter like so: 'protected baseUrl = "/my-base";'`);
        this.baseUrl = this.baseUrl ?? "/vix";
        const __dirname = path.dirname(fpath(this.routerImportUrl));
        const staticDir = path.resolve(__dirname, "../static");
        const entrypoint = path.resolve(staticDir, "index.html");
        const statics = staticRoute(staticDir, entrypoint);
        this.router.use(baseRedirectRoute(this.baseUrl, statics));
    }
    configureErrorHandling() {
        // Add this at the end to catch all Client Errors
        this.router.use(clientErrorRoute());
    }
}
