import fs from "node:fs";
import { INFO, OK, VERB, WARN } from "../util/logging.js";
import { parseExpressApp } from "./route-mapper.js";
function camelize(str) {
    return str
        .replace(/(?:^\w|[A-Z]|\b\w)/g, function (word, index) {
        return index === 0 ? word.toLowerCase() : word.toUpperCase();
    })
        .replace(/\s+/g, "")
        .replaceAll("-", "");
}
function loadContractName(contract) {
    const contractName = contract?.meta()?.vixContract;
    if (!contractName)
        throw Error("vixContract missing on route contract! Please add it to the schema meta!");
    return contractName;
}
function loadContractImport(contract) {
    const contractImport = contract?.meta()?.vixContractExport;
    if (!contractImport)
        throw Error("vixContractexport missing on route contract! Please add it to the schema meta!");
    return contractImport;
}
function loadRouteContracts(routeMetadata) {
    const imports = [];
    if (!routeMetadata?.vixContracts)
        return { imports, contracts: undefined };
    const vixContracts = routeMetadata.vixContracts;
    for (const vc in vixContracts) {
        const contract = vixContracts[vc];
        if (!contract)
            continue;
        imports.push(loadContractImport(contract));
    }
    return { imports, contracts: vixContracts };
}
function loadContractItems(contract) {
    const contractName = !!contract ? loadContractName(contract) : "";
    const contractImport = !!contract ? loadContractImport(contract) : "";
    const contractArg = camelize(contractName);
    return [contractName, contractImport, contractArg];
}
function getApiRoute(route, requestName, authLocked, subpath, contracts) {
    var requestNameSection = requestName;
    var subpathSection = `subpath: "${subpath}"`;
    const [reqContractName, reqContractImport, reqContractArg] = loadContractItems(contracts?.reqBodyContract);
    const [resContractName, resContractImport, resContractArg] = loadContractItems(contracts?.resBodyContract);
    const resBodyCast = !resContractName ? "" : `<${resContractImport}["${resContractName}"]>`;
    var jsonSection = !!reqContractName ? `json: ${reqContractArg}` : "";
    var functionArgs = {};
    var methodSection = `method:"${route.method.toUpperCase()}"`;
    var jsonifySection = "jsonify: true";
    if (route.pathParams.length > 0) {
        var normalizedSubpath = subpath;
        for (const p of route.pathParams) {
            normalizedSubpath = normalizedSubpath.replace(`:${p.name}`, "${" + p.name + "}");
            functionArgs[p.name] = "string";
        }
        subpathSection = "subpath: `" + normalizedSubpath + "`";
    }
    if (!!reqContractName)
        functionArgs[reqContractArg] = `${reqContractImport}["${reqContractName}"]`;
    const requestType = authLocked ? "authenticatedApiRequest" : "apiRequest";
    const argsSection = Object.keys(functionArgs).map((k) => `${k}: ${functionArgs[k]}`);
    const args = [subpathSection, methodSection, jsonSection, jsonifySection].filter((s) => !!s);
    return `
  export const ${requestNameSection} = (${argsSection}) =>
    ${requestType}${resBodyCast}({${args.join(", ")}});`;
}
export async function generateApi(app) {
    const routes = parseExpressApp(app);
    const filteredRoutes = routes.filter((route) => {
        if (!route.path.startsWith("/api")) {
            WARN("WARN", `Ignoring ${route.path}`);
            return false;
        }
        return true;
    });
    const apiLines = ["/* ----- Vix Auto Generated Routes ----- */"];
    const contractImports = new Set();
    for (const route of filteredRoutes) {
        const method = route.method ?? "all";
        const subpath = `/${route.path.split("/").slice(2).join("/")}`;
        var routePath = route.path;
        route.pathParams.forEach((p) => (routePath = routePath.replaceAll(`:${p.name}`, p.name)));
        const subpathSplit = camelize(method + routePath).split("/");
        subpathSplit.splice(1, 1);
        const subpathRequest = subpathSplit.join("");
        const authLocked = !!route?.routeMetadata?.authType;
        const { imports, contracts } = loadRouteContracts(route?.routeMetadata);
        imports.forEach((contractImport) => contractImports.add(contractImport));
        VERB("ROUTE", `Generating route for ${route.path}`);
        apiLines.push(getApiRoute(route, subpathRequest, authLocked, subpath, contracts));
    }
    const generatedLine = apiLines.shift();
    const apiRequestImport = `import { apiRequest } from "@dunemask` + `/vix/bridge";`; // TSC Alias will replace this if we don't split it up....
    const importLine = `import { ${Array.from(contractImports.values()).join(", ")} } from "@vix/ContractTypes";
${apiRequestImport}
import { authenticatedApiRequest } from "./requests";`;
    const apiImportedLines = [generatedLine, importLine, ...apiLines];
    INFO("INFO", "Saving to file....");
    fs.writeFileSync("src/util/api/GeneratedRequests.ts", apiImportedLines.join("\n"));
    OK("DONE", "DONE!");
    return apiImportedLines.join("\n");
}
