/**
 * Parses an Express app and generates list of routes with meta data
 *
 * @param app The Express app reference. Must be used after all routes have been attached
 * @returns List of routes for this express app with meta-data that has been picked up
 */
export const parseExpressApp = (app) => {
    return new ExpressPathParser(app).appPaths;
};
class ExpressPathParser {
    _appPaths;
    get appPaths() {
        return this._appPaths;
    }
    constructor(app) {
        this._appPaths = [];
        const router = app._router || app.router;
        if (!router)
            return;
        for (const layer of router.stack)
            this.traverse("", layer, []);
    }
    /**
     * Recursive traversal method for the express router and middleware tree.
     *
     * @param path The current path segment that we have traversed so far
     * @param layer The current 'layer' of the router tree
     * @param keys The keys for the parameter's in the current path branch of the traversal
     * @returns void - base case saves result to internal object
     */
    traverse = (path, layer, keys) => {
        keys = [...keys, ...layer.keys]; // Mismatch in provided express type
        if (layer.name === "router" && !!layer.handle?.stack) {
            for (const l of layer.handle.stack) {
                this.traverse(`${path || ""}/${pathRegexParser(layer.regexp, layer.keys)}`, l, keys);
            }
        }
        if (!layer.route || !layer.route.stack.length)
            return;
        this._appPaths.push(this.parseRouteLayer(layer, keys, path));
    };
    /**
     * Parses a route object. Route objects are the leafs of an express router tree
     *
     * @param layer The layer of this route object - represents the stack of middleware and other meta data
     * @param keys The full set of keys for this particular route
     * @param basePath The base path as it was initial declared for this route
     * @returns A ExpressPath object holding the meta data for a given route
     */
    parseRouteLayer = (layer, keys, basePath) => {
        if (!layer.route)
            throw Error("Can only parse router paths!");
        const lastRequestHandler = layer.route.stack[layer.route.stack.length - 1];
        const pathParams = keys.map((key) => ({ name: key.name, in: "path", required: !key.optional }));
        const filtered = layer.route?.stack.filter((element) => (element?.handle).routeMetadata);
        const path = (basePath + layer.route?.path).replace(/\/{2,}/gi, "/");
        if (filtered.length === 0)
            return { path, pathParams, method: lastRequestHandler.method };
        let metadata = {};
        for (const routerLayer of filtered) {
            const layerMetadata = routerLayer.handle.routeMetadata;
            const existingKeys = Object.keys(metadata);
            if (!!layerMetadata)
                for (const k of existingKeys)
                    if (k in layerMetadata)
                        console.warn(duplicateWarning(k));
            metadata = { ...metadata, ...layerMetadata };
        }
        return { path, pathParams, method: lastRequestHandler.method, routeMetadata: metadata };
    };
}
const duplicateWarning = (k) => `duplicate keys '${k}' will only use the last value on the router stack!`;
/** Parses an express layer's regex and converts it to the original format seen in code.
 *
 * @param layerRegexPath The layer's regex pattern
 * @param keys The keys that represent the layer's path parameters
 * @returns The path string for that layer
 * Code inspired and modify from:
 * https://github.com/expressjs/express/issues/3308#issuecomment-300957572
 */
const pathRegexParser = (layerRegexPath, keys) => {
    if (typeof layerRegexPath === "string")
        return layerRegexPath;
    const mappedPath = keys && keys.length ? mapKeysToPath(layerRegexPath, keys) : "";
    const match = layerRegexPath
        .toString()
        .replace("\\/?", "")
        .replace("(?=\\/|$)", "$")
        .match(/^\/\^((?:\\[.*+?^${}()|[\]\\/]|[^.*+?^${}()|[\]\\/])*)\$\//);
    if (match)
        return match[1].replace(/\\(.)/g, "$1").slice(1);
    if (mappedPath)
        return mappedPath.slice(1);
    return layerRegexPath.toString();
};
/**
 * Map's the keys/path variables to the regex inside a given path
 *
 * @param layerRegexPath The regex for a react router with path parameters
 * @param keys The keys that represent the path parameters
 * @returns The regex for a path variable converted to original string on the express route
 */
const mapKeysToPath = (layerRegexPath, keys) => {
    if (!keys || keys.length === 0)
        throw Error("must include atleast one key to map");
    let convertedSubPath = layerRegexPath.toString();
    for (const key of keys) {
        if (key.optional)
            convertedSubPath = convertedSubPath.replace("(?:\\/([^\\/]+?))?\\", `/:${key.name}?`);
        else
            convertedSubPath = convertedSubPath.replace("(?:([^\\/]+?))", `:${key.name}`);
    }
    return convertedSubPath
        .replace("/?(?=\\/|$)/i", "")
        .replace("/^", "")
        .replace(/\\/gi, "")
        .replace(/\/{2,}/gi, "/");
};
export function routePreviews(app, cb) {
    const routes = parseExpressApp(app);
    const maxMethodLength = Math.max(...routes.map(({ method }) => (method ?? "").length));
    const routePreviews = [];
    for (const route of routes) {
        const routeMethod = (route.method ?? "ALL").toUpperCase();
        const methodDisplay = new Array(maxMethodLength)
            .fill(0)
            .map((_v, i) => (i < routeMethod.length ? routeMethod.charAt(i) : " "))
            .join("");
        routePreviews.push(cb ? cb(route, methodDisplay) : `${methodDisplay} ${route.path}`);
    }
    return routePreviews;
}
export const onlyForTesting = { pathRegexParser, mapKeysToPath };
