export var ResourcePolicyFilters;
(function (ResourcePolicyFilters) {
    ResourcePolicyFilters["ANY"] = "any";
    ResourcePolicyFilters["ALL"] = "all";
})(ResourcePolicyFilters || (ResourcePolicyFilters = {}));
export var Policy;
(function (Policy) {
    Policy["ROOT"] = "root";
    Policy["CREATE"] = "create";
    Policy["READ"] = "read";
    Policy["WRITE"] = "write";
    Policy["DELETE"] = "delete";
    Policy["ANY"] = "*";
})(Policy || (Policy = {}));
const InvalidPolicy = Error(`Invalid policy! Valid options are ${Object.values(Policy).join(", ")}`);
export class ResourcePolicy {
    static parsePolicy(policy) {
        const parts = policy.split(".");
        if (parts.length !== 2)
            throw Error("Invalid resource policy!");
        if (!Object.values(Policy).includes(parts[1]))
            throw InvalidPolicy;
        return [parts[0], parts[1]];
    }
    static parseResourcePolicies(resources) {
        const resourceArr = Array.isArray(resources) ? resources : [resources];
        return resourceArr.map(function convertToPolicy(r) {
            if (r instanceof ResourcePolicy)
                return r;
            return new ResourcePolicy(r);
        });
    }
    static multiAuthorizedTo(policies, required, filter = ResourcePolicyFilters.ANY) {
        const existingPolicies = ResourcePolicy.parseResourcePolicies(policies);
        const requiredPolicies = ResourcePolicy.parseResourcePolicies(required);
        const isAllFilter = filter === ResourcePolicyFilters.ALL;
        const isAnyFilter = filter === ResourcePolicyFilters.ANY;
        for (const e of existingPolicies) {
            for (const r of requiredPolicies) {
                const isAuthorized = e.authorizedTo(r);
                if (!isAuthorized && isAllFilter)
                    return false;
                else if (isAuthorized && isAnyFilter)
                    return true;
            }
        }
        return isAllFilter;
    }
    constructor(resource, policy) {
        [this.resource, this.policy] = !!policy
            ? [resource, policy]
            : ResourcePolicy.parsePolicy(resource);
    }
    toString() {
        return `${this.resource}.${this.policy}`;
    }
    authorizedTo(policy) {
        if (this.resource !== policy.resource)
            return false; // Resource doesn't match, no authority
        if (this.policy === Policy.ROOT)
            return true; // User has root policy of this resource
        if (policy.policy === Policy.ANY)
            return true; // Resource has any policy
        if (this.policy === policy.policy)
            return true; // Policy Identically matches
        // Match additional "sub policies" (If a user can create, they can also do _____)
        if (this.policy === Policy.WRITE && policy.policy === Policy.READ)
            return true;
        if (this.policy === Policy.CREATE)
            return [Policy.WRITE, Policy.READ].includes(policy.policy);
        if (this.policy === Policy.DELETE)
            return [Policy.WRITE, Policy.READ].includes(policy.policy);
        return false;
    }
    static asStrings(resourcePolicies) {
        return resourcePolicies.map((rp) => rp.toString());
    }
    static multiple(...resourcePolicies) {
        return ResourcePolicy.parseResourcePolicies(resourcePolicies);
    }
}
